@@ -18,26 +18,11 @@
 #include <cuda_fp16.h>
 #include <cuda_runtime.h>
 
+#include "flashinfer/fp4_layout.cuh"
 #include "tensorrt_llm/common/quantization.h"
 
 namespace tensorrt_llm {
-
-enum class QuantizationSFLayout {
-  // Block scale factors are stored in swizzled layout for cutlass FP4 kernel. Scale factor
-  // blocks are organized in 512-byte blocks in global memory, with each block having 128x4 FP8
-  // values. The SF matrix dimensions are therefore padded - rows to the nearest multiple of 128 and
-  // columns to the nearest multiple of 4.
-  //
-  // The scale factor block rows map to data block rows in an interleaved pattern:
-  // For a scale factor row 'i', it maps to data block row: (i % 4) * 32 + (i / 4)
-  // Column 'j' in the scale factor block corresponds to scaling the j-th block in the data tensor.
-  SWIZZLED_128x4,
-  SWIZZLED_8x4,
-
-  // Block scale factors are stored in linear layout (row-major). This is used in some trtllm-gen
-  // kernels standard.
-  LINEAR
-};
+using flashinfer::QuantizationSFLayout;
 
 // This denotes the input and output data types of the block scale quantization.
 enum class BlockScaleQuantizationType {
@@ -77,9 +62,14 @@ void invokeFP4Quantization(int b, int m, int n, T const* input, float const* glo
                            QuantizationSFLayout layout, int multiProcessorCount,
                            bool enable_pdl = false, cudaStream_t stream = 0);
 
-void invokeBlockScaleInterleave(int b, int m, int m_padded, int n, int n_padded,
-                                uint8_t const* SFIn, uint8_t* SFOutput, int multiProcessorCount,
-                                cudaStream_t stream = 0);
+template <typename T>
+void invokeSiluAndMulNVFP4Quantization(void* output, void* output_scale, void* input,
+                                       void* input_global_scale, void* mask, bool use_silu_and_mul,
+                                       int m_topk, int k, int n_experts, cudaStream_t stream);
+
+template <typename T>
+void invokeBlockScaleInterleave(int b, int m, int m_padded, int n, int n_padded, T const* SFIn,
+                                T* SFOutput, int multiProcessorCount, cudaStream_t stream = 0);
 
 void invokeBlockScaleInterleaveReverse(int b, int m, int n, uint8_t const* SFIn, uint8_t* SFOutput,
                                        int multiProcessorCount, cudaStream_t stream = 0);
