@@ -3,7 +3,8 @@ from typing import Optional
 
 import torch
 
-from ..jit import get_cudnn_fmha_gen_module
+from ..api_logging import flashinfer_api
+from .utils import get_cudnn_fmha_gen_module
 
 try:
     import cudnn
@@ -252,6 +253,7 @@ def _batch_decode_with_kv_cache(
     return out
 
 
+@flashinfer_api
 def cudnn_batch_decode_with_kv_cache(
     q: torch.Tensor,
     k_cache: torch.Tensor,
