@@ -19,6 +19,7 @@ from .core import (
     gen_gemm_sm100_module_cutlass_fp4,
     gen_gemm_sm120_module_cutlass_fp4,
     gen_gemm_sm100_module_cutlass_fp8,
+    gen_gemm_sm100_module_cutlass_bf16,
     gen_gemm_sm100_module,
     gen_gemm_sm120_module,
     gen_trtllm_gen_gemm_module,
@@ -27,12 +28,14 @@ from .core import (
     gen_gemm_sm90_module,
 )
 from .deepgemm import gen_deepgemm_sm100_module
+from .fp8_blockscale import gen_fp8_blockscale_gemm_sm90_module
 
 __all__ = [
     "gen_gemm_module",
     "gen_gemm_sm100_module_cutlass_fp4",
     "gen_gemm_sm120_module_cutlass_fp4",
     "gen_gemm_sm100_module_cutlass_fp8",
+    "gen_gemm_sm100_module_cutlass_bf16",
     "gen_gemm_sm100_module",
     "gen_gemm_sm120_module",
     "gen_trtllm_gen_gemm_module",
@@ -40,4 +43,5 @@ __all__ = [
     "gen_tgv_gemm_sm10x_module",
     "gen_gemm_sm90_module",
     "gen_deepgemm_sm100_module",
+    "gen_fp8_blockscale_gemm_sm90_module",
 ]
