@@ -19,6 +19,7 @@ from typing import Optional
 
 import torch
 
+from .api_logging import flashinfer_api
 from .jit.norm import gen_norm_module
 from .utils import device_support_pdl, register_custom_op, register_fake_op
 
@@ -28,6 +29,7 @@ def get_norm_module():
     return gen_norm_module().build_and_load()
 
 
+@flashinfer_api
 def rmsnorm(
     input: torch.Tensor,
     weight: torch.Tensor,
@@ -90,6 +92,59 @@ def _rmsnorm_fake(
     pass
 
 
+@flashinfer_api
+@register_custom_op("flashinfer::rmsnorm_quant", mutates_args=("out",))
+def rmsnorm_quant(
+    out: torch.Tensor,
+    input: torch.Tensor,
+    weight: torch.Tensor,
+    scale: float,
+    eps: float = 1e-6,
+    enable_pdl: Optional[bool] = None,
+) -> torch.Tensor:
+    r"""Root mean square normalization.
+
+    ``out[i] = (input[i] / RMS(input)) * weight[i]``
+
+    Parameters
+    ----------
+    out: torch.Tensor
+        The output tensor, will quantize the output to the dtype of this tensor.
+    input: torch.Tensor
+        Input tensor, 2D shape (batch_size, hidden_size).
+    weight: torch.Tensor
+        Weight tensor, shape (hidden_size,).
+    scale: float
+        Scale factor for quantization.
+    eps: float
+        Epsilon for numerical stability.
+    enable_pdl: bool
+        Whether to enable `programmatic dependent launch
+        <https://docs.nvidia.com/cuda/cuda-c-programming-guide/index.html#programmatic-dependent-launch-and-synchronization>`_
+
+    Returns
+    -------
+    output: torch.Tensor
+        Normalized tensor, 2D shape (batch_size, hidden_size).
+    """
+    if enable_pdl is None:
+        enable_pdl = device_support_pdl(input.device)
+    get_norm_module().rmsnorm_quant(out, input, weight, scale, eps, enable_pdl)
+
+
+@register_fake_op("flashinfer::rmsnorm_quant")
+def _rmsnorm_quant_fake(
+    out: torch.Tensor,
+    input: torch.Tensor,
+    weight: torch.Tensor,
+    scale: float,
+    eps: float,
+    enable_pdl: Optional[bool],
+) -> None:
+    pass
+
+
+@flashinfer_api
 @register_custom_op("flashinfer::fused_add_rmsnorm", mutates_args=("input", "residual"))
 def fused_add_rmsnorm(
     input: torch.Tensor,
@@ -136,6 +191,66 @@ def _fused_add_rmsnorm_fake(
     pass
 
 
+@flashinfer_api
+@register_custom_op(
+    "flashinfer::fused_add_rmsnorm_quant", mutates_args=("out", "residual")
+)
+def fused_add_rmsnorm_quant(
+    out: torch.Tensor,
+    input: torch.Tensor,
+    residual: torch.Tensor,
+    weight: torch.Tensor,
+    scale: float,
+    eps: float = 1e-6,
+    enable_pdl: Optional[bool] = None,
+) -> None:
+    r"""Fused add root mean square normalization.
+
+    Step 1:
+    ``residual[i] += input[i]``
+
+    Step 2:
+    ``input[i] = (residual[i] / RMS(residual)) * weight[i]``
+
+    Parameters
+    ----------
+    out: torch.Tensor
+        The output tensor, will quantize the output to the dtype of this tensor.
+    input: torch.Tensor
+        Input tensor, shape (batch_size, hidden_size).
+    residual: torch.Tensor
+        Residual tensor, shape (batch_size, hidden_size).
+    weight: torch.Tensor
+        Weight tensor, shape (hidden_size,).
+    scale: float
+        Scale factor for quantization.
+    eps: float
+        Epsilon for numerical stability.
+    enable_pdl: bool
+        Whether to enable `programmatic dependent launch
+        <https://docs.nvidia.com/cuda/cuda-c-programming-guide/index.html#programmatic-dependent-launch-and-synchronization>`_
+    """
+    if enable_pdl is None:
+        enable_pdl = device_support_pdl(input.device)
+    get_norm_module().fused_add_rmsnorm_quant(
+        out, input, residual, weight, scale, eps, enable_pdl
+    )
+
+
+@register_fake_op("flashinfer::fused_add_rmsnorm_quant")
+def _fused_add_rmsnorm_quant_fake(
+    out: torch.Tensor,
+    input: torch.Tensor,
+    residual: torch.Tensor,
+    weight: torch.Tensor,
+    scale: float,
+    eps: float = 1e-6,
+    enable_pdl: Optional[bool] = None,
+) -> None:
+    pass
+
+
+@flashinfer_api
 def gemma_rmsnorm(
     input: torch.Tensor,
     weight: torch.Tensor,
@@ -198,6 +313,7 @@ def _gemma_rmsnorm_fake(
     pass
 
 
+@flashinfer_api
 @register_custom_op(
     "flashinfer::gemma_fused_add_rmsnorm", mutates_args=("input", "residual")
 )
@@ -244,3 +360,54 @@ def _gemma_fused_add_rmsnorm_fake(
     enable_pdl: Optional[bool] = None,
 ) -> None:
     pass
+
+
+@flashinfer_api
+@register_custom_op("flashinfer::layernorm", mutates_args=())
+def layernorm(
+    input: torch.Tensor,
+    gemma: torch.Tensor,
+    beta: torch.Tensor,
+    eps: float = 1e-6,
+) -> torch.Tensor:
+    r"""Layer normalization.
+    Parameters
+    ----------
+    input: torch.Tensor
+        Input tensor, shape (batch_size, hidden_size). Need to be bfloat16.
+    gemma: torch.Tensor
+        Gemma tensor, shape (hidden_size,). Need to be float32.
+    beta: torch.Tensor
+        Beta tensor, shape (hidden_size,). Need to be float32.
+    eps: float
+        Epsilon for numerical stability.
+
+    Returns
+    -------
+    output: torch.Tensor
+        Layer Normalized tensor, shape (batch_size, hidden_size). Same dtype as input.
+    """
+    out = torch.empty_like(input)
+    get_norm_module().layernorm(out, input, gemma, beta, eps)
+    return out
+
+
+@register_fake_op("flashinfer::layernorm")
+def _layernorm_fake(
+    input: torch.Tensor,
+    gemma: torch.Tensor,
+    beta: torch.Tensor,
+    eps: float = 1e-6,
+) -> torch.Tensor:
+    b, k = input.shape
+    return input.new_empty([b, k])
+
+
+# CuTe-DSL fused RMSNorm + FP4 Quantization kernels
+# These require CuTe-DSL to be available and SM100+ (Blackwell) GPUs
+try:
+    from .cute_dsl import rmsnorm_fp4quant, add_rmsnorm_fp4quant
+except ImportError:
+    # CuTe-DSL not available
+    rmsnorm_fp4quant = None  # type: ignore[misc,assignment]
+    add_rmsnorm_fp4quant = None  # type: ignore[misc,assignment]
