@@ -16,30 +16,37 @@ limitations under the License.
 
 import functools
 from types import SimpleNamespace
-from typing import Optional, Union
-
+from typing import Optional, Tuple, Union
 import torch
 
-from .jit import JitSpec
-from .jit import env as jit_env
-from .jit import gen_jit_spec
+from .api_logging import flashinfer_api
+from .jit.sampling import gen_sampling_module
 from .utils import (
     _get_cache_buf,
     device_support_pdl,
+    get_default_generators,
     register_custom_op,
     register_fake_op,
 )
 
 
-def gen_sampling_module() -> JitSpec:
-    return gen_jit_spec(
-        "sampling",
-        [
-            jit_env.FLASHINFER_CSRC_DIR / "sampling.cu",
-            jit_env.FLASHINFER_CSRC_DIR / "renorm.cu",
-            jit_env.FLASHINFER_CSRC_DIR / "flashinfer_sampling_ops.cu",
-        ],
+def get_seed_and_offset(
+    increment: int,
+    generator: Optional[torch.Generator] = None,
+    device: Optional[torch.device] = None,
+) -> Tuple[int, int]:
+    if generator is None:
+        generator = get_default_generators(device)
+    # add mutex if multi-trheading needed
+    state = generator.get_state()
+    seed, offset = state.view(torch.int64)
+    offset += (increment + 3) // 4 * 4
+    generator.set_state(
+        torch.tensor(
+            [seed, offset], dtype=torch.int64, device=torch.device("cpu")
+        ).view(torch.uint8)
     )
+    return int(seed), int(offset)
 
 
 @functools.cache
@@ -59,7 +66,7 @@ def get_sampling_module():
         maybe_temperature_arr = (
             maybe_temperature_arr.float() if maybe_temperature_arr is not None else None
         )
-        module.softmax.default(
+        module.softmax(
             workspace_buffer,
             logits,
             probs,
@@ -86,19 +93,27 @@ def get_sampling_module():
         indices: Optional[torch.Tensor],
         deterministic: bool,
         generator: Optional[torch.Generator],
+        seed: Optional[int] = None,
+        offset: Optional[int] = None,
     ) -> torch.Tensor:
         device = logits.device
         # TODO: support more data types in logits to avoid conversion
         # to float32
         logits = logits.float()
         batch_size = indices.size(0) if indices is not None else logits.size(0)
-        samples = torch.empty(batch_size, dtype=torch.int32, device=device)
-        module.sampling_from_logits.default(
+        out_dtype = indices.dtype if indices is not None else torch.int32
+        samples = torch.empty(batch_size, dtype=out_dtype, device=device)
+        if seed is None or offset is None:
+            seed, offset = get_seed_and_offset(
+                batch_size * logits.size(1), generator, device
+            )
+        module.sampling_from_logits(
             logits,
             samples,
             indices,
             deterministic,
-            generator,
+            seed,
+            offset,
         )
         return samples
 
@@ -110,7 +125,8 @@ def get_sampling_module():
         generator: Optional[torch.Generator],
     ) -> torch.Tensor:
         batch_size = indices.size(0) if indices is not None else logits.size(0)
-        return torch.empty(batch_size, dtype=torch.int32, device=logits.device)
+        out_dtype = indices.dtype if indices is not None else torch.int32
+        return torch.empty(batch_size, dtype=out_dtype, device=logits.device)
 
     # torch library for sampling_from_probs
 
@@ -120,17 +136,23 @@ def get_sampling_module():
         indices: Optional[torch.Tensor],
         deterministic: bool,
         generator: Optional[torch.Generator],
+        seed: Optional[int] = None,
+        offset: Optional[int] = None,
     ) -> torch.Tensor:
         device = probs.device
         probs = probs.float()
         batch_size = indices.size(0) if indices is not None else probs.size(0)
-        samples = torch.empty(batch_size, dtype=torch.int32, device=device)
-        module.sampling_from_probs.default(
+        out_dtype = indices.dtype if indices is not None else torch.int32
+        samples = torch.empty(batch_size, dtype=out_dtype, device=device)
+        if seed is None or offset is None:
+            seed, offset = get_seed_and_offset(batch_size, generator, device)
+        module.sampling_from_probs(
             probs,
             samples,
             indices,
             deterministic,
-            generator,
+            seed,
+            offset,
         )
         return samples
 
@@ -144,7 +166,8 @@ def get_sampling_module():
         generator: Optional[torch.Generator],
     ) -> torch.Tensor:
         batch_size = indices.size(0) if indices is not None else probs.size(0)
-        return torch.empty(batch_size, dtype=torch.int32, device=probs.device)
+        out_dtype = indices.dtype if indices is not None else torch.int32
+        return torch.empty(batch_size, dtype=out_dtype, device=probs.device)
 
     # torch library for top_p_sampling_from_probs
 
@@ -156,6 +179,8 @@ def get_sampling_module():
         top_p_val: float,
         deterministic: bool,
         generator: Optional[torch.Generator],
+        seed: Optional[int] = None,
+        offset: Optional[int] = None,
     ) -> torch.Tensor:
         device = probs.device
         probs = probs.float()
@@ -163,15 +188,19 @@ def get_sampling_module():
             maybe_top_p_arr.float() if maybe_top_p_arr is not None else None
         )
         batch_size = indices.size(0) if indices is not None else probs.size(0)
-        samples = torch.empty(batch_size, dtype=torch.int32, device=device)
-        module.top_p_sampling_from_probs.default(
+        out_dtype = indices.dtype if indices is not None else torch.int32
+        samples = torch.empty(batch_size, dtype=out_dtype, device=device)
+        if seed is None or offset is None:
+            seed, offset = get_seed_and_offset(batch_size * 32, generator, device)
+        module.top_p_sampling_from_probs(
             probs,
             samples,
             indices,
             maybe_top_p_arr,
             top_p_val,
             deterministic,
-            generator,
+            seed,
+            offset,
         )
         return samples
 
@@ -184,7 +213,9 @@ def get_sampling_module():
         deterministic: bool,
         generator: Optional[torch.Generator],
     ) -> torch.Tensor:
-        sample = torch.empty(probs.size(0), dtype=torch.int32, device=probs.device)
+        batch_size = indices.size(0) if indices is not None else probs.size(0)
+        out_dtype = indices.dtype if indices is not None else torch.int32
+        sample = torch.empty(batch_size, dtype=out_dtype, device=probs.device)
         return sample
 
     # torch library for top_k_sampling_from_probs
@@ -197,20 +228,26 @@ def get_sampling_module():
         top_k_val: int,
         deterministic: bool,
         generator: Optional[torch.Generator],
+        seed: Optional[int] = None,
+        offset: Optional[int] = None,
     ) -> torch.Tensor:
         device = probs.device
         probs = probs.float()
         batch_size = indices.size(0) if indices is not None else probs.size(0)
         maybe_top_k_arr = maybe_top_k_arr.int() if maybe_top_k_arr is not None else None
-        samples = torch.empty(batch_size, dtype=torch.int32, device=device)
-        module.top_k_sampling_from_probs.default(
+        out_dtype = indices.dtype if indices is not None else torch.int32
+        samples = torch.empty(batch_size, dtype=out_dtype, device=device)
+        if seed is None or offset is None:
+            seed, offset = get_seed_and_offset(batch_size * 32, generator, device)
+        module.top_k_sampling_from_probs(
             probs,
             samples,
             indices,
             maybe_top_k_arr,
             top_k_val,
             deterministic,
-            generator,
+            seed,
+            offset,
         )
         return samples
 
@@ -224,7 +261,8 @@ def get_sampling_module():
         generator: Optional[torch.Generator],
     ) -> torch.Tensor:
         batch_size = indices.size(0) if indices is not None else probs.size(0)
-        sample = torch.empty(batch_size, dtype=torch.int32, device=probs.device)
+        out_dtype = indices.dtype if indices is not None else torch.int32
+        sample = torch.empty(batch_size, dtype=out_dtype, device=probs.device)
         return sample
 
     # torch library for min_p_sampling_from_probs
@@ -237,6 +275,8 @@ def get_sampling_module():
         min_p_val: float,
         deterministic: bool,
         generator: Optional[torch.Generator],
+        seed: Optional[int] = None,
+        offset: Optional[int] = None,
     ) -> torch.Tensor:
         device = probs.device
         probs = probs.float()
@@ -244,15 +284,19 @@ def get_sampling_module():
             maybe_min_p_arr.float() if maybe_min_p_arr is not None else None
         )
         batch_size = indices.size(0) if indices is not None else probs.size(0)
-        samples = torch.empty(batch_size, dtype=torch.int32, device=device)
-        module.min_p_sampling_from_probs.default(
+        out_dtype = indices.dtype if indices is not None else torch.int32
+        samples = torch.empty(batch_size, dtype=out_dtype, device=device)
+        if seed is None or offset is None:
+            seed, offset = get_seed_and_offset(batch_size, generator, device)
+        module.min_p_sampling_from_probs(
             probs,
             samples,
             indices,
             maybe_min_p_arr,
             min_p_val,
             deterministic,
-            generator,
+            seed,
+            offset,
         )
         return samples
 
@@ -268,6 +312,8 @@ def get_sampling_module():
         top_p_val: float,
         deterministic: bool,
         generator: Optional[torch.Generator],
+        seed: Optional[int] = None,
+        offset: Optional[int] = None,
     ) -> torch.Tensor:
         device = probs.device
         probs = probs.float()
@@ -276,8 +322,11 @@ def get_sampling_module():
             maybe_top_p_arr.float() if maybe_top_p_arr is not None else None
         )
         batch_size = indices.size(0) if indices is not None else probs.size(0)
-        samples = torch.empty(batch_size, dtype=torch.int32, device=device)
-        module.top_k_top_p_sampling_from_probs.default(
+        out_dtype = indices.dtype if indices is not None else torch.int32
+        samples = torch.empty(batch_size, dtype=out_dtype, device=device)
+        if seed is None or offset is None:
+            seed, offset = get_seed_and_offset(batch_size * 32, generator, device)
+        module.top_k_top_p_sampling_from_probs(
             probs,
             samples,
             indices,
@@ -286,7 +335,8 @@ def get_sampling_module():
             maybe_top_p_arr,
             top_p_val,
             deterministic,
-            generator,
+            seed,
+            offset,
         )
         return samples
 
@@ -302,7 +352,8 @@ def get_sampling_module():
         generator: Optional[torch.Generator],
     ) -> torch.Tensor:
         batch_size = indices.size(0) if indices is not None else probs.size(0)
-        sample = torch.empty(batch_size, dtype=torch.int32, device=probs.device)
+        out_dtype = indices.dtype if indices is not None else torch.int32
+        sample = torch.empty(batch_size, dtype=out_dtype, device=probs.device)
         return sample
 
     # torch library for top_p_renorm_probs
@@ -318,7 +369,7 @@ def get_sampling_module():
             maybe_top_p_arr.float() if maybe_top_p_arr is not None else None
         )
         renorm_probs = torch.empty_like(probs)
-        module.top_p_renorm_probs.default(
+        module.top_p_renorm_probs(
             probs,
             renorm_probs,
             maybe_top_p_arr,
@@ -336,20 +387,27 @@ def get_sampling_module():
 
     # torch library for top_k_renorm_probs
 
-    @register_custom_op("flashinfer::top_k_renorm_probs", mutates_args=())
+    @register_custom_op(
+        "flashinfer::top_k_renorm_probs", mutates_args=("row_states_buffer",)
+    )
     def top_k_renorm_probs(
         probs: torch.Tensor,
         maybe_top_k_arr: Optional[torch.Tensor],
         top_k_val: int,
+        row_states_buffer: torch.Tensor,
     ) -> torch.Tensor:
-        probs = probs.float()
+        # Support FP32, FP16, BF16
+        assert probs.dtype in [torch.float32, torch.float16, torch.bfloat16], (
+            f"Unsupported dtype {probs.dtype}, expected float32, float16, or bfloat16"
+        )
         maybe_top_k_arr = maybe_top_k_arr.int() if maybe_top_k_arr is not None else None
         renorm_probs = torch.empty_like(probs)
-        module.top_k_renorm_probs.default(
+        module.top_k_renorm_probs(
             probs,
             renorm_probs,
             maybe_top_k_arr,
             top_k_val,
+            row_states_buffer,
         )
         return renorm_probs
 
@@ -358,25 +416,34 @@ def get_sampling_module():
         probs: torch.Tensor,
         maybe_top_k_arr: Optional[torch.Tensor],
         top_k_val: int,
+        row_states_buffer: torch.Tensor,
     ) -> torch.Tensor:
         return torch.empty_like(probs)
 
     # torch library for top_k_mask_logits
 
-    @register_custom_op("flashinfer::top_k_mask_logits", mutates_args=())
+    @register_custom_op(
+        "flashinfer::top_k_mask_logits", mutates_args=("row_states_buffer",)
+    )
     def top_k_mask_logits(
         logits: torch.Tensor,
         maybe_top_k_arr: Optional[torch.Tensor],
         top_k_val: int,
+        row_states_buffer: torch.Tensor,
     ) -> torch.Tensor:
-        logits = logits.float()
+        # Support FP32, FP16, BF16
+        assert logits.dtype in [torch.float32, torch.float16, torch.bfloat16], (
+            f"Unsupported dtype {logits.dtype}, expected float32, float16, or bfloat16"
+        )
         maybe_top_k_arr = maybe_top_k_arr.int() if maybe_top_k_arr is not None else None
         mask_logits = torch.empty_like(logits)
-        module.top_k_mask_logits.default(
+
+        module.top_k_mask_logits(
             logits,
             mask_logits,
             maybe_top_k_arr,
             top_k_val,
+            row_states_buffer,
         )
         return mask_logits
 
@@ -385,6 +452,7 @@ def get_sampling_module():
         logits: torch.Tensor,
         maybe_top_k_arr: Optional[torch.Tensor],
         top_k_val: int,
+        row_states_buffer: torch.Tensor,
     ) -> torch.Tensor:
         return torch.empty_like(logits)
 
@@ -405,6 +473,8 @@ def get_sampling_module():
         output_emitted_draft_token_num: torch.Tensor,
         deterministic: bool,
         generator: Optional[torch.Generator],
+        seed: Optional[int] = None,
+        offset: Optional[int] = None,
     ) -> torch.Tensor:
         device = draft_probs.device
         draft_probs = draft_probs.float()
@@ -414,7 +484,11 @@ def get_sampling_module():
         output_emitted_draft_token_num = output_emitted_draft_token_num.int()
         b, n = draft_token_ids.shape
         output_token_ids = torch.empty((b, n + 1), dtype=torch.int32, device=device)
-        module.chain_speculative_sampling.default(
+        if seed is None or offset is None:
+            seed, offset = get_seed_and_offset(
+                draft_probs.size(0) * (draft_probs.size(1) + 1), generator, device
+            )
+        module.chain_speculative_sampling(
             draft_probs,
             draft_token_ids,
             target_probs,
@@ -422,7 +496,8 @@ def get_sampling_module():
             output_accepted_token_num,
             output_emitted_draft_token_num,
             deterministic,
-            generator,
+            seed,
+            offset,
         )
         return output_token_ids
 
@@ -463,6 +538,7 @@ def _to_tensor_scalar_tuple(x):
         return (None, x)
 
 
+@flashinfer_api
 def softmax(
     logits: torch.Tensor,
     temperature: Optional[Union[torch.Tensor, float]] = None,
@@ -520,12 +596,15 @@ def softmax(
     )
 
 
+@flashinfer_api
 def sampling_from_logits(
     logits: torch.Tensor,
     indices: Optional[torch.Tensor] = None,
     deterministic: bool = True,
     generator: Optional[torch.Generator] = None,
     check_nan: bool = False,
+    seed: Optional[int] = None,
+    offset: Optional[int] = None,
 ) -> torch.Tensor:
     r"""Fused GPU kernel for category sampling from logits. It's equivalent to sampling
     from :attr:`logits` after applying softmax.
@@ -537,10 +616,12 @@ def sampling_from_logits(
         shape should be ``(unique_batch_size, num_classes)`` where unique_batch_size is the number of unique
         probability distributions.
     indices: Optional[torch.Tensor]
-        Optional indices tensor of shape ``(batch_size,)`` that maps each output to a row in logits.
+        Optional indices tensor of shape ``(batch_size,)``, dtype ``torch.int32`` or ``torch.int64``
+        that maps each output to a row in logits. The output tensor will have the same dtype as indices.
         For example, if indices[i] = j, then the i-th output will be sampled from logits[j].
         This allows reusing the same probability distribution for multiple outputs.
-        If indices is not provided, the i-th output will be sampled from the i-th row of logits.
+        If indices is not provided, the i-th output will be sampled from the i-th row of logits
+        and output dtype defaults to ``torch.int32``.
     deterministic: bool
         Since the sampling doesn't use cub's BlockScan, the sampling is deterministic. We keep this
         argument for compatibility with other sampling functions.
@@ -548,6 +629,10 @@ def sampling_from_logits(
         A random number generator for the operation.
     check_nan: bool
         Whether to check nan in :attr:`logits`, default is ``False``.
+    seed: Optional[int]
+        seed value to use for the rng during the sampling operation.
+    offset: Optional[int]
+        offset value to use for the rng during the sampling operation.
     Returns
     -------
     samples: torch.Tensor
@@ -574,16 +659,19 @@ def sampling_from_logits(
         if torch.any(torch.isnan(logits)):
             raise ValueError("Input logits contains NaN.")
     return get_sampling_module().sampling_from_logits(
-        logits, indices, deterministic, generator
+        logits, indices, deterministic, generator, seed, offset
     )
 
 
+@flashinfer_api
 def sampling_from_probs(
     probs: torch.Tensor,
     indices: Optional[torch.Tensor] = None,
     deterministic: bool = True,
     generator: Optional[torch.Generator] = None,
     check_nan: bool = False,
+    seed: Optional[int] = None,
+    offset: Optional[int] = None,
 ) -> torch.Tensor:
     r"""Fused GPU kernel for category sampling from probabilities.
 
@@ -595,16 +683,22 @@ def sampling_from_probs(
         shape should be ``(unique_batch_size, num_classes)`` where unique_batch_size is the number of unique
         probability distributions.
     indices: Optional[torch.Tensor]
-        Optional indices tensor of shape ``(batch_size,)`` that maps each output to a row in probs.
+        Optional indices tensor of shape ``(batch_size,)``, dtype ``torch.int32`` or ``torch.int64``
+        that maps each output to a row in probs. The output tensor will have the same dtype as indices.
         For example, if indices[i] = j, then the i-th output will be sampled from probs[j].
         This allows reusing the same probability distribution for multiple outputs.
-        If indices is not provided, the i-th output will be sampled from the i-th row of probs.
+        If indices is not provided, the i-th output will be sampled from the i-th row of probs
+        and output dtype defaults to ``torch.int32``.
     deterministic: bool
         Whether to use deterministic kernel implementation, default is ``True``.
     generator: Optional[torch.Generator]
         A random number generator for the operation.
     check_nan: bool
         Whether to check nan in :attr:`probs`, default is ``False``.
+    seed: Optional[int]
+        seed value to use for the rng during the sampling operation.
+    offset: Optional[int]
+        offset value to use for the rng during the sampling operation.
 
     Returns
     -------
@@ -638,10 +732,11 @@ def sampling_from_probs(
         if torch.any(torch.isnan(probs)):
             raise ValueError("Input probs contains NaN.")
     return get_sampling_module().sampling_from_probs(
-        probs, indices, deterministic, generator
+        probs, indices, deterministic, generator, seed, offset
     )
 
 
+@flashinfer_api
 def top_p_sampling_from_probs(
     probs: torch.Tensor,
     top_p: Union[torch.Tensor, float],
@@ -649,6 +744,8 @@ def top_p_sampling_from_probs(
     deterministic: bool = True,
     generator: Optional[torch.Generator] = None,
     check_nan: bool = False,
+    seed: Optional[int] = None,
+    offset: Optional[int] = None,
 ) -> torch.Tensor:
     r"""Fused GPU kernel for top-p sampling (nucleus sampling) from probabilities,
     this operator implements GPU-based rejection sampling without explicit sorting.
@@ -665,20 +762,26 @@ def top_p_sampling_from_probs(
         shape should be ``(unique_batch_size, num_classes)`` where unique_batch_size is the number of unique
         probability distributions.
     top_p: Union[torch.Tensor, float]
-        Either a scalar or a tensor of shape ``(batch_size,)``, representing the threshold for top-p sampling.
-        If a scalar, the same threshold is used for all requests.
+        Either a float or a tensor of shape ``(batch_size,)``, representing the threshold for top-p sampling.
+        If a float, the same threshold is used for all requests.
         If a tensor, each request has its own threshold.
     indices: Optional[torch.Tensor]
-        Optional indices tensor of shape ``(batch_size,)`` that maps each output to a row in probs.
+        Optional indices tensor of shape ``(batch_size,)``, dtype ``torch.int32`` or ``torch.int64``
+        that maps each output to a row in probs. The output tensor will have the same dtype as indices.
         For example, if indices[i] = j, then the i-th output will be sampled from probs[j].
         This allows reusing the same probability distribution for multiple outputs.
-        If indices is not provided, the i-th output will be sampled from the i-th row of probs.
+        If indices is not provided, the i-th output will be sampled from the i-th row of probs
+        and output dtype defaults to ``torch.int32``.
     deterministic: bool
         Whether to use deterministic kernel implementation, default is ``True``.
     generator: Optional[torch.Generator]
         A random number generator for the operation.
     check_nan: bool
         Whether to check nan in :attr:`probs`, default is ``False``.
+    seed: Optional[int]
+        seed value to use for the rng during the sampling operation.
+    offset: Optional[int]
+        offset value to use for the rng during the sampling operation.
 
     Returns
     -------
@@ -720,10 +823,17 @@ def top_p_sampling_from_probs(
         if torch.any(torch.isnan(probs)):
             raise ValueError("Input probs contains NaN.")
     return get_sampling_module().top_p_sampling_from_probs(
-        probs, indices, *_to_tensor_scalar_tuple(top_p), deterministic, generator
+        probs,
+        indices,
+        *_to_tensor_scalar_tuple(top_p),
+        deterministic,
+        generator,
+        seed,
+        offset,
     )
 
 
+@flashinfer_api
 def top_k_sampling_from_probs(
     probs: torch.Tensor,
     top_k: Union[torch.Tensor, int],
@@ -731,6 +841,8 @@ def top_k_sampling_from_probs(
     deterministic: bool = True,
     generator: Optional[torch.Generator] = None,
     check_nan: bool = False,
+    seed: Optional[int] = None,
+    offset: Optional[int] = None,
 ) -> torch.Tensor:
     r"""Fused GPU kernel for top-k sampling from probabilities,
     this operator implements GPU-based rejection sampling without explicit sorting.
@@ -751,16 +863,22 @@ def top_k_sampling_from_probs(
         If a scalar, the same threshold is used for all requests.
         If a tensor, each request has its own threshold.
     indices: Optional[torch.Tensor]
-        Optional indices tensor of shape ``(batch_size,)`` that maps each output to a row in probs.
+        Optional indices tensor of shape ``(batch_size,)``, dtype ``torch.int32`` or ``torch.int64``
+        that maps each output to a row in probs. The output tensor will have the same dtype as indices.
         For example, if indices[i] = j, then the i-th output will be sampled from probs[j].
         This allows reusing the same probability distribution for multiple outputs.
-        If indices is not provided, the i-th output will be sampled from the i-th row of probs.
+        If indices is not provided, the i-th output will be sampled from the i-th row of probs
+        and output dtype defaults to ``torch.int32``.
     deterministic: bool
         Whether to use deterministic kernel implementation, default is ``True``.
     generator: Optional[torch.Generator]
         A random number generator for the operation.
     check_nan: bool
         Whether to check nan in :attr:`probs`, default is ``False``.
+    seed: Optional[int]
+        seed value to use for the rng during the sampling operation.
+    offset: Optional[int]
+        offset value to use for the rng during the sampling operation.
 
     Returns
     -------
@@ -802,10 +920,17 @@ def top_k_sampling_from_probs(
         if torch.any(torch.isnan(probs)):
             raise ValueError("Input probs contains NaN.")
     return get_sampling_module().top_k_sampling_from_probs(
-        probs, indices, *_to_tensor_scalar_tuple(top_k), deterministic, generator
+        probs,
+        indices,
+        *_to_tensor_scalar_tuple(top_k),
+        deterministic,
+        generator,
+        seed,
+        offset,
     )
 
 
+@flashinfer_api
 def min_p_sampling_from_probs(
     probs: torch.Tensor,
     min_p: Union[torch.Tensor, float],
@@ -813,6 +938,8 @@ def min_p_sampling_from_probs(
     deterministic: bool = True,
     generator: Optional[torch.Generator] = None,
     check_nan: bool = False,
+    seed: Optional[int] = None,
+    offset: Optional[int] = None,
 ) -> torch.Tensor:
     r"""Fused GPU kernel for `min_p sampling <https://arxiv.org/abs/2407.01082>`_ from probabilities,
 
@@ -834,16 +961,22 @@ def min_p_sampling_from_probs(
         If a scalar, the same threshold is used for all requests.
         If a tensor, each request has its own threshold.
     indices: Optional[torch.Tensor]
-        Optional indices tensor of shape ``(batch_size,)`` that maps each output to a row in probs.
+        Optional indices tensor of shape ``(batch_size,)``, dtype ``torch.int32`` or ``torch.int64``
+        that maps each output to a row in probs. The output tensor will have the same dtype as indices.
         For example, if indices[i] = j, then the i-th output will be sampled from probs[j].
         This allows reusing the same probability distribution for multiple outputs.
-        If indices is not provided, the i-th output will be sampled from the i-th row of probs.
+        If indices is not provided, the i-th output will be sampled from the i-th row of probs
+        and output dtype defaults to ``torch.int32``.
     deterministic: bool
         Whether to use deterministic kernel implementation, default is ``True``.
     generator: Optional[torch.Generator]
         A random number generator for the operation.
     check_nan: bool
         Whether to check nan in :attr:`probs`, default is ``False``.
+    seed: Optional[int]
+        seed value to use for the rng during the sampling operation.
+    offset: Optional[int]
+        offset value to use for the rng during the sampling operation.
 
     Returns
     -------
@@ -880,10 +1013,17 @@ def min_p_sampling_from_probs(
         if torch.any(torch.isnan(probs)):
             raise ValueError("Input probs contains NaN.")
     return get_sampling_module().min_p_sampling_from_probs(
-        probs, indices, *_to_tensor_scalar_tuple(min_p), deterministic, generator
+        probs,
+        indices,
+        *_to_tensor_scalar_tuple(min_p),
+        deterministic,
+        generator,
+        seed,
+        offset,
     )
 
 
+@flashinfer_api
 def top_k_top_p_sampling_from_logits(
     logits: torch.Tensor,
     top_k: Union[torch.Tensor, int],
@@ -893,6 +1033,8 @@ def top_k_top_p_sampling_from_logits(
     deterministic: bool = True,
     generator: Optional[torch.Generator] = None,
     check_nan: bool = False,
+    seed: Optional[int] = None,
+    offset: Optional[int] = None,
 ) -> torch.Tensor:
     r"""Fused GPU kernel for top-k and top-p sampling from pre-softmax logits,
 
@@ -918,10 +1060,12 @@ def top_k_top_p_sampling_from_logits(
         If a scalar, the same threshold is used for all requests.
         If a tensor, each request has its own threshold.
     indices: Optional[torch.Tensor]
-        Optional indices tensor of shape ``(batch_size,)`` that maps each output to a row in probs.
+        Optional indices tensor of shape ``(batch_size,)``, dtype ``torch.int32`` or ``torch.int64``
+        that maps each output to a row in probs. The output tensor will have the same dtype as indices.
         For example, if indices[i] = j, then the i-th output will be sampled from probs[j].
         This allows reusing the same probability distribution for multiple outputs.
-        If indices is not provided, the i-th output will be sampled from the i-th row of probs.
+        If indices is not provided, the i-th output will be sampled from the i-th row of probs
+        and output dtype defaults to ``torch.int32``.
     filter_apply_order: str
         The order of applying top-k and top-p sampling, should be either ``"top_k_first"`` or ``"joint"``.
         If ``"top_k_first"``, we first apply top-k filter, then apply top-p sampling on the top-k results.
@@ -932,6 +1076,10 @@ def top_k_top_p_sampling_from_logits(
         A random number generator for the operation.
     check_nan: bool
         Whether to check nan in :attr:`probs`, default is ``False``.
+    seed: Optional[int]
+        seed value to use for the rng during the sampling operation.
+    offset: Optional[int]
+        offset value to use for the rng during the sampling operation.
 
     Returns
     -------
@@ -986,6 +1134,8 @@ def top_k_top_p_sampling_from_logits(
             deterministic,
             check_nan=check_nan,
             generator=generator,
+            seed=seed,
+            offset=offset,
         )
     elif filter_apply_order == "joint":
         probs = torch.softmax(logits, dim=-1)
@@ -999,11 +1149,14 @@ def top_k_top_p_sampling_from_logits(
             *_to_tensor_scalar_tuple(top_p),
             deterministic,
             generator,
+            seed,
+            offset,
         )
     else:
         raise ValueError(f"Invalid filter_apply_order: {filter_apply_order}")
 
 
+@flashinfer_api
 def top_k_top_p_sampling_from_probs(
     probs: torch.Tensor,
     top_k: Union[torch.Tensor, int],
@@ -1013,6 +1166,8 @@ def top_k_top_p_sampling_from_probs(
     deterministic: bool = True,
     generator: Optional[torch.Generator] = None,
     check_nan: bool = False,
+    seed: Optional[int] = None,
+    offset: Optional[int] = None,
 ) -> torch.Tensor:
     r"""Fused GPU kernel for top-k and top-p sampling from probabilities,
 
@@ -1038,10 +1193,12 @@ def top_k_top_p_sampling_from_probs(
         If a scalar, the same threshold is used for all requests.
         If a tensor, each request has its own threshold.
     indices: Optional[torch.Tensor]
-        Optional indices tensor of shape ``(batch_size,)`` that maps each output to a row in probs.
+        Optional indices tensor of shape ``(batch_size,)``, dtype ``torch.int32`` or ``torch.int64``
+        that maps each output to a row in probs. The output tensor will have the same dtype as indices.
         For example, if indices[i] = j, then the i-th output will be sampled from probs[j].
         This allows reusing the same probability distribution for multiple outputs.
-        If indices is not provided, the i-th output will be sampled from the i-th row of probs.
+        If indices is not provided, the i-th output will be sampled from the i-th row of probs
+        and output dtype defaults to ``torch.int32``.
     filter_apply_order: str
         The order of applying top-k and top-p sampling, should be either ``"top_k_first"`` or ``"joint"``.
         If ``"top_k_first"``, we first apply top-k filter, then apply top-p sampling on the top-k results.
@@ -1052,6 +1209,10 @@ def top_k_top_p_sampling_from_probs(
         A random number generator for the operation.
     check_nan: bool
         Whether to check nan in :attr:`probs`, default is ``False``.
+    seed: Optional[int]
+        seed value to use for the rng during the sampling operation.
+    offset: Optional[int]
+        offset value to use for the rng during the sampling operation.
 
     Returns
     -------
@@ -1100,6 +1261,8 @@ def top_k_top_p_sampling_from_probs(
             deterministic,
             check_nan=check_nan,
             generator=generator,
+            seed=seed,
+            offset=offset,
         )
     elif filter_apply_order == "joint":
         if check_nan:
@@ -1112,11 +1275,14 @@ def top_k_top_p_sampling_from_probs(
             *_to_tensor_scalar_tuple(top_p),
             deterministic,
             generator,
+            seed,
+            offset,
         )
     else:
         raise ValueError(f"Invalid filter_apply_order: {filter_apply_order}")
 
 
+@flashinfer_api
 def top_p_renorm_probs(
     probs: torch.Tensor,
     top_p: Union[torch.Tensor, float],
@@ -1182,6 +1348,7 @@ def top_p_renorm_probs(
 top_p_renorm_prob = top_p_renorm_probs
 
 
+@flashinfer_api
 def top_k_renorm_probs(
     probs: torch.Tensor,
     top_k: Union[torch.Tensor, int],
@@ -1192,6 +1359,7 @@ def top_k_renorm_probs(
     ----------
     probs: torch.Tensor
         Probabilities, shape ``(batch_size, num_classes)``.
+        Supported dtypes: ``float32``, ``float16``, ``bfloat16``.
     top_k: Union[torch.Tensor, int]
         Either a scalar or a tensor of shape ``(batch_size,)``, representing the top-k threshold for for
         for re-normalizing probabilities, should be in ``(0, num_classes)``.
@@ -1203,6 +1371,7 @@ def top_k_renorm_probs(
     -------
     renorm_probs: torch.Tensor
         Renormalized probabilities, shape ``(batch_size, num_classes)``.
+        Same dtype as input ``probs``.
 
     Examples
     --------
@@ -1237,15 +1406,26 @@ def top_k_renorm_probs(
     top_k_sampling_from_probs
     sampling_from_probs
     top_p_renorm_probs
+    top_k : General-purpose top-k selection (returns indices and values)
     """
+    # Allocate row_states buffer for multi-CTA kernel (1MB is enough for any GPU)
+    buffer_bytes = 1024 * 1024  # 1MB
+    row_states_buffer = _get_cache_buf(
+        f"top_k_renorm_probs_row_states_{probs.device}",
+        buffer_bytes,
+        probs.device,
+        zero_init=True,
+    )
+
     return get_sampling_module().top_k_renorm_probs(
-        probs, *_to_tensor_scalar_tuple(top_k)
+        probs, *_to_tensor_scalar_tuple(top_k), row_states_buffer
     )
 
 
 top_k_renorm_prob = top_k_renorm_probs
 
 
+@flashinfer_api
 def top_k_mask_logits(
     logits: torch.Tensor, top_k: Union[torch.Tensor, int]
 ) -> torch.Tensor:
@@ -1255,6 +1435,7 @@ def top_k_mask_logits(
     ----------
     logits: torch.Tensor
         Logits before softmax, shape ``(batch_size, num_classes)``.
+        Supported dtypes: ``float32``, ``float16``, ``bfloat16``.
     top_k: Union[torch.Tensor, int]
         Either a scalar or a tensor of shape ``(batch_size,)``, representing the top-k threshold for for
         for masking logits, should be in ``(0, num_classes)``.
@@ -1266,6 +1447,7 @@ def top_k_mask_logits(
     -------
     masked_logits: torch.Tensor
         Masked logits, shape ``(batch_size, num_classes)``.
+        Same dtype as input ``logits``.
 
     Examples
     --------
@@ -1296,12 +1478,26 @@ def top_k_mask_logits(
     See Also
     --------
     top_k_renorm_probs
+    top_k : General-purpose top-k selection (returns indices and values)
     """
+    # Allocate row_states buffer for multi-CTA kernel (1MB is enough for any GPU)
+    buffer_bytes = 1024 * 1024  # 1MB
+    row_states_buffer = _get_cache_buf(
+        f"top_k_mask_logits_row_states_{logits.device}",
+        buffer_bytes,
+        logits.device,
+        zero_init=True,
+    )
+
+    # Note: row_states_buffer is zero-initialized on first allocation by _get_cache_buf
+    # Kernel will reset arrival_counter to 0 at the end of each launch
+
     return get_sampling_module().top_k_mask_logits(
-        logits, *_to_tensor_scalar_tuple(top_k)
+        logits, *_to_tensor_scalar_tuple(top_k), row_states_buffer
     )
 
 
+@flashinfer_api
 def chain_speculative_sampling(
     draft_probs,
     draft_token_ids,
@@ -1310,6 +1506,8 @@ def chain_speculative_sampling(
     maybe_output_emitted_draft_token_num: Optional[torch.Tensor] = None,
     deterministic: bool = True,
     generator: Optional[torch.Generator] = None,
+    seed: Optional[int] = None,
+    offset: Optional[int] = None,
 ) -> torch.Tensor:
     r"""Fused-GPU kernel for speculative sampling for sequence generation (proposed in
     paper `Accelerating Large Language Model Decoding with Speculative Sampling <https://arxiv.org/pdf/2302.01318>`_),
@@ -1345,6 +1543,10 @@ def chain_speculative_sampling(
         Whether to use deterministic kernel implementation, default is ``True``.
     generator: Optional[torch.Generator]
         A random number generator for the operation.
+    seed: Optional[int]
+        seed value to use for the rng during the sampling operation.
+    offset: Optional[int]
+        offset value to use for the rng during the sampling operation.
 
     Returns
     -------
@@ -1411,5 +1613,7 @@ def chain_speculative_sampling(
         output_emitted_draft_token_num,
         deterministic,
         generator,
+        seed,
+        offset,
     )
     return output_token_ids, output_accepted_token_num, output_emitted_draft_token_num
