@@ -138,7 +138,7 @@ class MemAllocatorHelper {
 
 ////////////////////////////////////////////////////////////////////////////////////////////////////
 
-inline int getNumSmemBitsPerElt(tg::Dtype dtype, tg::MmaKind mmaKind) {
+inline int getNumSmemBitsPerElt(tg::Dtype dtype, tg::MmaKind mmaKind, int mmaK) {
   if (mmaKind == tg::MmaKind::Auto) {
     throw std::runtime_error("mmaKind != tg::MmaKind::Auto");
   }
@@ -168,7 +168,8 @@ class KernelTraits {
                bool usePerTokenSfB, bool useTwoCtas, BiasType biasType)
       : mMmaKind{mmaKind},
         mFuseUtccpWithUtcmma{fuseUtccpWithUtcmma},
-        mUseMaxTmemOverlap{useMaxTmemOverlap} {
+        mUseMaxTmemOverlap{useMaxTmemOverlap},
+        mNumEpilogueWarps{numEpilogueWarps} {
     //
     // SMEM
     //
@@ -202,7 +203,7 @@ class KernelTraits {
       {
         // Number of bytes in load A shared memory.
         auto const numSmemBytesLoadA =
-            numStages * tileM * tileK * getNumSmemBitsPerElt(dtypeA, mMmaKind) / 8 /* bits */;
+            numStages * tileM * tileK * getNumSmemBitsPerElt(dtypeA, mMmaKind, mmaK) / 8 /* bits */;
         // Number of bytes for load A alignment for TMA load.
         auto const numBytesAlignmentLoadA = 1024;
         // loadA is already at first chunk. No need to reuse it.
@@ -218,7 +219,7 @@ class KernelTraits {
       {
         // Number of bytes in load B shared memory.
         auto const numSmemBytesLoadB = numStages * (useTwoCtas ? tileN / 2 : tileN) * tileK *
-                                       getNumSmemBitsPerElt(dtypeB, mMmaKind) / 8 /* bits */;
+                                       getNumSmemBitsPerElt(dtypeB, mMmaKind, mmaK) / 8 /* bits */;
         // Number of bytes for load B alignment for TMA load.
         auto const numBytesAlignmentLoadB = 1024;
         // No need to reuse the first chunk.
@@ -238,9 +239,9 @@ class KernelTraits {
       {
         // Number of bytes in save shuffled B in shared memory.
         auto const numSmemBytesLoadB =
-            numSlicesForSliceK > 1
-                ? numStages * tileN * tileK * getNumSmemBitsPerElt(dtypeB, mMmaKind) / 8 /* bits */
-                : 0;
+            numSlicesForSliceK > 1 ? numStages * tileN * tileK *
+                                         getNumSmemBitsPerElt(dtypeB, mMmaKind, mmaK) / 8 /* bits */
+                                   : 0;
         // Number of bytes for load B alignment for TMA load.
         auto const numBytesAlignmentLoadB = 1024;
         // No need to reuse the first chunk.
@@ -474,13 +475,19 @@ class KernelTraits {
         bool const useBlockScalingA = tg::dtypeIsBlockFmt(dtypeMmaA);
         // Are the block scales constant?
         bool const useConstSfA = useBlockScalingA && !tg::dtypeIsBlockFmt(dtypeA);
+        // Number elements per scaling factor.
+        int32_t const numEltsPerSf = useBlockScalingA ? tg::dtypeNumEltsPerSf(dtypeMmaA) : -1;
+        // TMEM cols group size in the K dimension.
+        int32_t kGroupSize = 4;
+        // Number of columns per stage.
+        int32_t const numColsPerStage =
+            useBlockScalingA ? ((tileK / (kGroupSize * numEltsPerSf)) *
+                                tg::getTmemColStridePerGroup(tileM, mmaK, kGroupSize))
+                             : 0;
         // Number of columns for scaling factors of A.
         auto const numTmemColsSfA =
-            useConstSfA
-                ? tg::roundUp((tileK / 64) * tg::getTmemColStridePerGroup(tileM, mmaK), 4)
-                : (useBlockScalingA ? ((tileK / 64) * tg::getTmemColStridePerGroup(tileM, mmaK)) *
-                                          (mFuseUtccpWithUtcmma ? 1 : numStages)
-                                    : 0);
+            useConstSfA ? tg::roundUp(numColsPerStage, 4)
+                        : (numColsPerStage * (mFuseUtccpWithUtcmma ? 1 : numStages));
         // Number of columns for Sf alignment.
         auto const numColsAlignmentSfA = 4;
         // No need to reuse TMEM.
@@ -499,13 +506,19 @@ class KernelTraits {
         bool const useBlockScalingB = tg::dtypeIsBlockFmt(dtypeMmaB);
         // Are the block scales constant?
         bool const useConstSfB = useBlockScalingB && !tg::dtypeIsBlockFmt(dtypeB);
+        // Number elements per scaling factor.
+        int32_t const numEltsPerSf = useBlockScalingB ? tg::dtypeNumEltsPerSf(dtypeMmaB) : -1;
+        // TMEM cols group size in the K dimension.
+        int32_t kGroupSize = 4;
+        // Number of columns per stage.
+        int32_t const numColsPerStage =
+            useBlockScalingB ? ((tileK / (kGroupSize * numEltsPerSf)) *
+                                tg::getTmemColStridePerGroup(tileN, mmaK, kGroupSize))
+                             : 0;
         // Number of columns for scaling factors of B.
         auto const numTmemColsSfB =
-            useConstSfB
-                ? tg::roundUp((tileK / 64) * tg::getTmemColStridePerGroup(tileN, mmaK), 4)
-                : (useBlockScalingB ? ((tileK / 64) * tg::getTmemColStridePerGroup(tileN, mmaK)) *
-                                          (mFuseUtccpWithUtcmma ? 1 : numStages)
-                                    : 0);
+            useConstSfB ? tg::roundUp(numColsPerStage, 4)
+                        : (numColsPerStage * (mFuseUtccpWithUtcmma ? 1 : numStages));
         // Number of columns for Sf alignment.
         auto const numColsAlignmentSfB = 4;
         // No need to reuse TMEM.
@@ -531,6 +544,8 @@ class KernelTraits {
   bool mFuseUtccpWithUtcmma;
   // Whether use the max TMEM overlap trick.
   bool mUseMaxTmemOverlap;
+  // The number of epilogue warps.
+  int32_t mNumEpilogueWarps;
   // Helper for SMEM allocation.
   MemAllocatorHelper mSmemAllocatorHelper;
   // Helper for TMEM allocation.
